"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateType = exports.generate = void 0;
var lodash_1 = require("lodash");
var index_1 = require("./index");
var AST_1 = require("./types/AST");
var utils_1 = require("./utils");
function generate(ast, options) {
    var _a;
    if (options === void 0) { options = index_1.DEFAULT_OPTIONS; }
    var named = declareNamedTypes(ast, options, ast.standaloneName);
    var interfaces = declareNamedInterfaces(ast, options, ast.standaloneName);
    var enums = declareEnums(ast, options);
    // @TODO: group by module
    var imports = generateImports((0, lodash_1.flatMap)(__spreadArray(__spreadArray(__spreadArray([], (0, lodash_1.flatMap)(named, function (n) { var _a; return (_a = n === null || n === void 0 ? void 0 : n.imports) !== null && _a !== void 0 ? _a : []; }), true), (0, lodash_1.flatMap)(interfaces, function (i) { var _a; return (_a = i === null || i === void 0 ? void 0 : i.imports) !== null && _a !== void 0 ? _a : []; }), true), [
        (_a = enums.imports) !== null && _a !== void 0 ? _a : []
    ], false)));
    return ([
        options.bannerComment,
        imports,
        named
            .map(function (n) { return n === null || n === void 0 ? void 0 : n.type; })
            .filter(Boolean)
            .join('\n\n'),
        interfaces
            .map(function (n) { return n === null || n === void 0 ? void 0 : n.type; })
            .filter(Boolean)
            .join('\n\n'),
        enums.type
    ]
        .filter(Boolean)
        .join('\n\n') + '\n'); // trailing newline
}
exports.generate = generate;
function generateImports(imported) {
    var typeImports = {};
    var valueImports = {};
    imported.forEach(function (t) {
        var _a, _b;
        var name = t.name;
        if (t.defaultExport) {
            name = "default as {t.name}";
        }
        if (t.typeOnly) {
            typeImports[t.module] = __spreadArray(__spreadArray([], ((_a = typeImports[t.module]) !== null && _a !== void 0 ? _a : []), true), [name], false);
        }
        else {
            valueImports[t.module] = __spreadArray(__spreadArray([], ((_b = valueImports[t.module]) !== null && _b !== void 0 ? _b : []), true), [name], false);
        }
    });
    var importStmts = [];
    (0, lodash_1.forEach)(typeImports, function (names, module) {
        importStmts.push("import type {".concat(names.join(", "), "} from \"").concat(module, "\""));
    });
    (0, lodash_1.forEach)(valueImports, function (names, module) {
        importStmts.push("import {".concat(names.join(", "), "} from \"").concat(module, "\""));
    });
    return importStmts;
}
function declareEnums(ast, options, processed) {
    if (processed === void 0) { processed = new Set(); }
    if (processed.has(ast)) {
        return { type: '' };
    }
    processed.add(ast);
    switch (ast.type) {
        case 'ENUM':
            return generateStandaloneEnum(ast, options);
        case 'ARRAY':
            return declareEnums(ast.params, options, processed);
        case 'UNION':
        case 'INTERSECTION':
            return (function () {
                var types = ast.params.map(function (ast) { return declareEnums(ast, options, processed); });
                var names = types.reduce(function (prev, _a) {
                    var type = _a.type;
                    return prev + type;
                }, '');
                return {
                    imports: (0, lodash_1.flatMap)(types.map(function (t) { var _a; return (_a = t.imports) !== null && _a !== void 0 ? _a : []; })),
                    type: names
                };
            })();
        case 'TUPLE':
            return (function () {
                var types = ast.params.map(function (ast) { return declareEnums(ast, options, processed); });
                if (ast.spreadParam) {
                    types.push(declareEnums(ast.spreadParam, options, processed));
                }
                var name = types.reduce(function (prev, _a) {
                    var type = _a.type;
                    return prev + type;
                }, '');
                return {
                    imports: (0, lodash_1.flatMap)(types.map(function (t) { var _a; return (_a = t.imports) !== null && _a !== void 0 ? _a : []; })),
                    type: name
                };
            })();
        case 'INTERFACE':
            return (function () {
                var types = getSuperTypesAndParams(ast).map(function (ast) { return declareEnums(ast, options, processed); });
                var name = types.reduce(function (prev, _a) {
                    var type = _a.type;
                    return prev + type;
                }, '');
                return {
                    imports: (0, lodash_1.flatMap)(types.map(function (t) { var _a; return (_a = t.imports) !== null && _a !== void 0 ? _a : []; })),
                    type: name
                };
            })();
        default:
            return { type: '' };
    }
}
function declareNamedInterfaces(ast, options, rootASTName, processed) {
    if (processed === void 0) { processed = new Set(); }
    if (processed.has(ast)) {
        return [{ type: '' }];
    }
    processed.add(ast);
    switch (ast.type) {
        case 'ARRAY':
            return declareNamedInterfaces(ast.params, options, rootASTName, processed);
        case 'INTERFACE':
            return (0, lodash_1.flatMap)(__spreadArray([
                (0, AST_1.hasStandaloneName)(ast) && (ast.standaloneName === rootASTName || options.declareExternallyReferenced)
                    ? [generateStandaloneInterface(ast, options)]
                    : []
            ], getSuperTypesAndParams(ast).map(function (ast) { return declareNamedInterfaces(ast, options, rootASTName, processed); }), true));
        case 'INTERSECTION':
        case 'TUPLE':
        case 'UNION':
            var types = (0, lodash_1.flatMap)(ast.params, function (_) { return declareNamedInterfaces(_, options, rootASTName, processed); });
            if (ast.type === 'TUPLE' && ast.spreadParam) {
                types.push.apply(types, declareNamedInterfaces(ast.spreadParam, options, rootASTName, processed));
            }
            return types;
    }
    return [{ type: '' }];
}
function declareNamedTypes(ast, options, rootASTName, processed) {
    if (processed === void 0) { processed = new Set(); }
    if (processed.has(ast)) {
        return [{ type: '' }];
    }
    processed.add(ast);
    switch (ast.type) {
        case 'ARRAY':
            return __spreadArray(__spreadArray([], declareNamedTypes(ast.params, options, rootASTName, processed), true), [
                (0, AST_1.hasStandaloneName)(ast) ? generateStandaloneType(ast, options) : undefined
            ], false);
        case 'ENUM':
            return [{ type: '' }];
        case 'INTERFACE':
            return (0, lodash_1.flatMap)(getSuperTypesAndParams(ast), function (ast) {
                return ast.standaloneName === rootASTName || options.declareExternallyReferenced
                    ? declareNamedTypes(ast, options, rootASTName, processed)
                    : [];
            });
        case 'INTERSECTION':
        case 'TUPLE':
        case 'UNION':
            return __spreadArray(__spreadArray([
                (0, AST_1.hasStandaloneName)(ast) ? generateStandaloneType(ast, options) : undefined
            ], (0, lodash_1.flatMap)(ast.params, function (ast) { return declareNamedTypes(ast, options, rootASTName, processed); }), true), ('spreadParam' in ast && ast.spreadParam
                ? declareNamedTypes(ast.spreadParam, options, rootASTName, processed)
                : []), true);
        default:
            if ((0, AST_1.hasStandaloneName)(ast)) {
                return [generateStandaloneType(ast, options)];
            }
            return [{ type: '' }];
    }
}
function generateTypeUnmemoized(ast, options) {
    var type = generateRawType(ast, options);
    if (options.strictIndexSignatures && ast.keyName === '[k: string]') {
        return { type: "".concat(type.type, " | undefined"), imports: type.imports };
    }
    return type;
}
exports.generateType = (0, lodash_1.memoize)(generateTypeUnmemoized);
function generateRawType(ast, options) {
    (0, utils_1.log)('magenta', 'generator', ast);
    if ((0, AST_1.hasStandaloneName)(ast)) {
        return { type: (0, utils_1.toSafeString)(ast.standaloneName) };
    }
    switch (ast.type) {
        case 'ANY':
            return { type: 'any' };
        case 'ARRAY':
            return (function () {
                var _a = (0, exports.generateType)(ast.params, options), type = _a.type, imports = _a.imports;
                return { type: type.endsWith('"') ? '(' + type + ')[]' : type + '[]', imports: imports };
            })();
        case 'BOOLEAN':
            return { type: 'boolean' };
        case 'INTERFACE':
            return generateInterface(ast, options);
        case 'INTERSECTION':
            return generateSetOperation(ast, options);
        case 'LITERAL':
            return { type: JSON.stringify(ast.params) };
        case 'NEVER':
            return { type: 'never' };
        case 'NUMBER':
            return { type: 'number' };
        case 'NULL':
            return { type: 'null' };
        case 'OBJECT':
            return { type: 'object' };
        case 'REFERENCE':
            return { type: ast.params };
        case 'STRING':
            return { type: 'string' };
        case 'TUPLE':
            return (function () {
                var minItems = ast.minItems;
                var maxItems = ast.maxItems || -1;
                var spreadParam = ast.spreadParam;
                var astParams = __spreadArray([], ast.params, true);
                if (minItems > 0 && minItems > astParams.length && ast.spreadParam === undefined) {
                    // this is a valid state, and JSONSchema doesn't care about the item type
                    if (maxItems < 0) {
                        // no max items and no spread param, so just spread any
                        spreadParam = options.unknownAny ? AST_1.T_UNKNOWN : AST_1.T_ANY;
                    }
                }
                if (maxItems > astParams.length && ast.spreadParam === undefined) {
                    // this is a valid state, and JSONSchema doesn't care about the item type
                    // fill the tuple with any elements
                    for (var i = astParams.length; i < maxItems; i += 1) {
                        astParams.push(options.unknownAny ? AST_1.T_UNKNOWN : AST_1.T_ANY);
                    }
                }
                function addSpreadParam(params) {
                    if (spreadParam) {
                        var spread = '...(' + (0, exports.generateType)(spreadParam, options).type + ')[]';
                        params.push(spread);
                    }
                    return params;
                }
                function paramsToString(params) {
                    return '[' + params.join(', ') + ']';
                }
                var paramsList = astParams.map(function (param) { return (0, exports.generateType)(param, options); });
                var imports = (0, lodash_1.flatMap)(paramsList.map(function (p) { var _a; return (_a = p.imports) !== null && _a !== void 0 ? _a : []; }));
                if (paramsList.length > minItems) {
                    /*
          if there are more items than the min, we return a union of tuples instead of
          using the optional element operator. This is done because it is more typesafe.
          
          // optional element operator
          type A = [string, string?, string?]
          const a: A = ['a', undefined, 'c'] // no error
          
          // union of tuples
          type B = [string] | [string, string] | [string, string, string]
          const b: B = ['a', undefined, 'c'] // TS error
          */
                    var cumulativeParamsList = paramsList.map(function (p) { return p.type; }).slice(0, minItems);
                    var typesToUnion = [];
                    if (cumulativeParamsList.length > 0) {
                        // actually has minItems, so add the initial state
                        typesToUnion.push(paramsToString(cumulativeParamsList));
                    }
                    else {
                        // no minItems means it's acceptable to have an empty tuple type
                        typesToUnion.push(paramsToString([]));
                    }
                    for (var i = minItems; i < paramsList.length; i += 1) {
                        cumulativeParamsList.push(paramsList[i].type);
                        if (i === paramsList.length - 1) {
                            // only the last item in the union should have the spread parameter
                            addSpreadParam(cumulativeParamsList);
                        }
                        typesToUnion.push(paramsToString(cumulativeParamsList));
                    }
                    return { type: typesToUnion.join('|'), imports: imports };
                }
                // no max items so only need to return one type
                return {
                    type: paramsToString(addSpreadParam(paramsList.map(function (p) { return p.type; }))),
                    imports: imports
                };
            })();
        case 'UNION':
            return generateSetOperation(ast, options);
        case 'UNKNOWN':
            return { type: 'unknown' };
        case 'CUSTOM_TYPE':
            if (ast.params.module) {
                return {
                    type: ast.params.name,
                    imports: [{ module: ast.params.module, name: ast.params.name, defaultExport: ast.params.defaultExport }]
                };
            }
            return { type: ast.params.name };
    }
}
/**
 * Generate a Union or Intersection
 */
function generateSetOperation(ast, options) {
    var members = ast.params.map(function (_) { return (0, exports.generateType)(_, options); });
    var separator = ast.type === 'UNION' ? '|' : '&';
    return members.length === 1
        ? members[0]
        : {
            type: '(' + members.map(function (m) { return m.type; }).join(' ' + separator + ' ') + ')',
            imports: (0, lodash_1.flatMap)(members.map(function (m) { var _a; return (_a = m.imports) !== null && _a !== void 0 ? _a : []; }))
        };
}
function generateInterface(ast, options) {
    var fields = ast.params
        .filter(function (_) { return !_.isPatternProperty && !_.isUnreachableDefinition; })
        .map(function (_a) {
        var isRequired = _a.isRequired, keyName = _a.keyName, ast = _a.ast;
        return [isRequired, keyName, ast, (0, exports.generateType)(ast, options)];
    });
    return {
        imports: (0, lodash_1.flatMap)(fields, function (_a) {
            var _1 = _a[0], _2 = _a[1], _3 = _a[2], imports = _a[3].imports;
            return imports !== null && imports !== void 0 ? imports : [];
        }),
        type: "{" +
            '\n' +
            fields
                .map(function (_a) {
                var isRequired = _a[0], keyName = _a[1], ast = _a[2], type = _a[3].type;
                return ((0, AST_1.hasComment)(ast) && !ast.standaloneName ? generateComment(ast.comment) + '\n' : '') +
                    escapeKeyName(keyName) +
                    (isRequired ? '' : '?') +
                    ': ' +
                    ((0, AST_1.hasStandaloneName)(ast) ? (0, utils_1.toSafeString)(type) : type);
            })
                .join('\n') +
            '\n' +
            '}'
    };
}
function generateComment(comment) {
    return __spreadArray(__spreadArray(['/**'], comment.split('\n').map(function (_) { return ' * ' + _; }), true), [' */'], false).join('\n');
}
function generateStandaloneEnum(ast, options) {
    var types = ast.params.map(function (_a) {
        var ast = _a.ast, keyName = _a.keyName;
        return (__assign({ keyName: keyName }, (0, exports.generateType)(ast, options)));
    });
    var names = types.map(function (t) { return t.keyName + ' = ' + t.type; });
    var imports = (0, lodash_1.flatMap)(types, function (_a) {
        var imports = _a.imports;
        return imports !== null && imports !== void 0 ? imports : [];
    });
    return {
        imports: imports,
        type: ((0, AST_1.hasComment)(ast) ? generateComment(ast.comment) + '\n' : '') +
            'export ' +
            (options.enableConstEnums ? 'const ' : '') +
            "enum ".concat((0, utils_1.toSafeString)(ast.standaloneName), " {") +
            '\n' +
            names.join(',\n') +
            '\n' +
            '}'
    };
}
function generateStandaloneInterface(ast, options) {
    var iface = generateInterface(ast, options);
    return {
        imports: iface.imports,
        type: ((0, AST_1.hasComment)(ast) ? generateComment(ast.comment) + '\n' : '') +
            "export interface ".concat((0, utils_1.toSafeString)(ast.standaloneName), " ") +
            (ast.superTypes.length > 0
                ? "extends ".concat(ast.superTypes.map(function (superType) { return (0, utils_1.toSafeString)(superType.standaloneName); }).join(', '), " ")
                : '') +
            iface.type
    };
}
function generateStandaloneType(ast, options) {
    var _a = (0, exports.generateType)((0, lodash_1.omit)(ast, 'standaloneName') /* TODO */, options), type = _a.type, imports = _a.imports;
    var exportedType = ((0, AST_1.hasComment)(ast) ? generateComment(ast.comment) + '\n' : '') +
        "export type ".concat((0, utils_1.toSafeString)(ast.standaloneName), " = ").concat(type);
    return {
        imports: imports,
        type: type ? exportedType : ''
    };
}
function escapeKeyName(keyName) {
    if (keyName.length && /[A-Za-z_$]/.test(keyName.charAt(0)) && /^[\w$]+$/.test(keyName)) {
        return keyName;
    }
    if (keyName === '[k: string]') {
        return keyName;
    }
    return JSON.stringify(keyName);
}
function getSuperTypesAndParams(ast) {
    return ast.params.map(function (param) { return param.ast; }).concat(ast.superTypes);
}
//# sourceMappingURL=generator.js.map